import logging

from glyphspkg.filenames import userNameToFileName
from glyphspkg.paths import build_output_file_path
from glyphspkg.plist import parse_plist_from_path, save_to_plist_path
from pathlib import Path
from typing import Any, Dict, List, Optional, Union


logger = logging.getLogger(__name__)


def package_to_single(input_path: Path, output_path: Optional[Path] = None) -> Path:
    # The main dict
    glyphs_file = convert_fontinfo(input_path)

    # Glyph order, is used to read individual glyphs files
    glyph_order = convert_order(input_path)

    # All glyphs files derived from the Glyph order list
    glyphs = []
    for glyph_name in glyph_order:
        file_name = userNameToFileName(glyph_name)
        file_path = input_path / "glyphs" / f"{file_name}.glyph"
        if not file_path.is_file():
            logger.warning(
                f"Glyph file not found for glyph '{glyph_name}': "
                f"{file_name}, glyph will be missing in converted file."
            )
            continue

        glyph = parse_plist_from_path(file_path)
        glyphs.append(glyph)

    glyphs_file["glyphs"] = glyphs

    # UIState, current display strings
    uistate = convert_uistate(input_path)
    if uistate:
        # Why the different key casing?
        glyphs_file["DisplayStrings"] = uistate["displayStrings"]

    output_file_path = build_output_file_path(input_path, output_path)
    if input_path == output_file_path:
        logger.error(f"Saving would overwrite the input file {input_path}")
        raise FileExistsError

    logger.info(f"Saving: {output_file_path}")
    save_to_plist_path(glyphs_file, output_file_path)
    return output_file_path


def convert_fontinfo(input_path: Path) -> Union[Dict[Any, Any], List[Any]]:
    return parse_plist_from_path(input_path / "fontinfo.plist")


def convert_order(input_path: Path) -> Union[Dict[Any, Any], List[Any]]:
    return parse_plist_from_path(input_path / "order.plist")


def convert_uistate(input_path: Path) -> Union[Dict[Any, Any], List[Any]]:
    uistate_path = input_path / "UIState.plist"
    if not uistate_path.is_file():
        return {}

    return parse_plist_from_path(uistate_path)



"""
Command line interface for compilation/decompilation
"""

import argparse
from glyphspkg.package_to_single import package_to_single
from glyphspkg.single_to_package import single_to_package
from pathlib import Path


def main() -> None:
    parser = argparse.ArgumentParser()

    # parser.add_argument(
    #     "-i",
    #     "--in-place",
    #     action="store_true",
    #     default=False,
    #     help="Convert the file in place, i.e. remove the input file.",
    # )
    parser.add_argument(
        "-o",
        "--output",
        dest="output_path",
        type=str,
        help=(
            "Output path for converted files. If omitted, the file is saved "
            "next to the input file."
        ),
    )
    parser.add_argument(
        "glyphsfile",
        type=str,
        nargs="+",
        help="Path to the .glyphs file or .glyphspackage dir to be converted.",
    )

    arguments = parser.parse_args()

    if not arguments:
        parser.print_help()
        return

    if arguments.output_path is None:
        output_path = None
    else:
        output_path = Path(arguments.output_path)
        if not output_path.is_dir() and len(arguments.glyphsfile) > 1:
            print(
                "ERROR: If you specify a full output file path, you must only "
                "supply one input file path, or else the output file will be "
                "overwritten with each input file."
            )
            return

    for inp in arguments.glyphsfile:
        input_path = Path(inp)
        print("Reading:", input_path)
        if input_path.is_dir():
            # The file is a package, convert to single
            output_file_path = package_to_single(input_path, output_path)
            print(f"Saved <{output_file_path}>.")
        else:
            # The file is a single file, convert to package
            output_file_path = single_to_package(input_path, output_path)
            print(f"Saved <{output_file_path}>.")



from pathlib import Path
from typing import Optional


def build_output_file_path(
    input_path: Path, output_path: Optional[Path] = None, target_suffix: str = ".glyphs"
) -> Path:
    file_name = input_path.with_suffix(target_suffix).name
    if output_path is None:
        # No path was specified, save next to original file
        output_file_path = input_path.parent / file_name
    elif output_path.is_dir():
        # Output directory was specified, save to original file name with
        # changed suffix in new dir
        output_file_path = output_path / file_name
    else:
        # Full path with file name was specified, save there
        output_file_path = output_path

    return output_file_path


def rmdir(directory: Path):
    """Recursively remove files and directories from a directory, then remove the
    directory itself.

    Args:
        directory (Path): The directory to remove.
    """
    for item in directory.iterdir():
        if item.is_dir():
            rmdir(item)
        else:
            item.unlink()
    directory.rmdir()



import codecs
import openstep_plist

from pathlib import Path
from typing import Any, Dict, List, Union


def parse_plist_from_path(
    plist_path: Path,
) -> Union[Dict[Any, Any], List[Any]]:
    with codecs.open(str(plist_path), "rb", "utf-8") as f:
        d = f.read()
    return parse(d)


def save_to_plist_path(obj: Union[Dict[Any, Any], List[Any]], plist_path: Path) -> None:
    with codecs.open(str(plist_path), "wb", "utf-8") as f:
        openstep_plist.dump(
            obj,
            f,
            unicode_escape=False,
            indent=0,
            single_line_tuples=True,
            # TODO: The released version 0.3.1 of openstep_plist doesn't have the
            # ``escape_newlines`` argument
            # escape_newlines=False,
        )


# From glyphsLib.parser
# https://github.com/googlefonts/glyphsLib
# Licensed under Apache-2.0


def parse(d: str) -> Union[Dict[Any, Any], List[Any]]:
    try:
        d = _fl7_format_clean(d)
        result: Union[Dict[Any, Any], List[Any]] = openstep_plist.loads(
            d, use_numbers=True
        )
    except openstep_plist.parser.ParseError as e:
        raise ValueError("Failed to parse file") from e
    return result


def _fl7_format_clean(d: str) -> str:
    """
    FontLab 7 glyphs source format exports include a final closing semicolon.
    This method removes the semicolon before passing the string to the parser.
    """
    # see https://github.com/googlefonts/fontmake/issues/806
    return d.rstrip(";\n")



import logging

from glyphspkg.filenames import userNameToFileName
from glyphspkg.paths import build_output_file_path, rmdir
from glyphspkg.plist import parse_plist_from_path, save_to_plist_path
from pathlib import Path
from typing import Any, Dict, List, Optional, Union


logger = logging.getLogger(__name__)


def single_to_package(input_path: Path, output_path: Optional[Path] = None) -> Path:
    output_file_path = build_output_file_path(input_path, output_path, ".glyphspackage")
    if input_path == output_file_path:
        logger.error(f"Saving would overwrite the input file {input_path}")
        raise FileExistsError

    # The main dict
    glyphs_file = parse_plist_from_path(input_path)

    logger.info(f"Saving: {output_file_path}")

    # Remove the directory if it exists
    if output_file_path.exists():
        logger.info(f"Output file exists, overwriting: <{output_file_path}>")
        if output_file_path.is_dir():
            rmdir(output_file_path)
        else:
            output_file_path.unlink()

    # Make the directory structure
    (output_file_path / "glyphs").mkdir(parents=True)

    # Glyphs
    glyphs = glyphs_file.get("glyphs", [])
    convert_glyphs(glyphs, output_file_path)

    uistate = {"displayStrings": glyphs_file.get("DisplayStrings", [])}
    convert_uistate(uistate, output_file_path)

    # Fontinfo
    # The font info is the original glyphs file structure with the separately written
    # keys removed
    for key in ("DisplayStrings", "glyphs"):
        if key in glyphs_file:
            del glyphs_file[key]
    convert_fontinfo(glyphs_file, output_file_path)

    return output_file_path


def convert_fontinfo(
    fontinfo: Union[Dict[Any, Any], List[Any]],
    output_path: Path,
) -> None:
    save_to_plist_path(fontinfo, output_path / "fontinfo.plist")


def convert_glyphs(
    glyphs: Union[Dict[Any, Any], List[Any]],
    output_path: Path,
) -> None:
    order = []
    for glyph in glyphs:
        name = glyph["glyphname"]
        order.append(name)
        save_to_plist_path(
            glyph, output_path / "glyphs" / f"{userNameToFileName(name)}.glyph"
        )
    convert_order(order, output_path)


def convert_order(
    order: Union[Dict[Any, Any], List[Any]],
    output_path: Path,
) -> None:
    save_to_plist_path(order, output_path / "order.plist")


def convert_uistate(
    uistate: Union[Dict[Any, Any], List[Any]],
    output_path: Path,
) -> None:
    save_to_plist_path(uistate, output_path / "UIState.plist")



import pytest
import unittest

from glyphspkg.package_to_single import package_to_single
from glyphspkg.paths import rmdir
from pathlib import Path


def get_mono_path():
    return (
        Path(__file__).parent / "data" / "JetBrainsMono-Italic.glyphspackage"
    )


def get_tmp_path():
    return Path(__file__).parent / "data" / "tmp"


class PackageToSingleText(unittest.TestCase):
    def test_same_dir(self):
        package_to_single(get_mono_path())
        out_file = get_mono_path().with_suffix(".glyphs")
        assert out_file.is_file()
        # TODO: Check against original file?
        out_file.unlink()

    def test_with_output_dir(self):
        out_dir = get_tmp_path()
        out_dir.mkdir(exist_ok=True)
        package_to_single(get_mono_path(), out_dir)
        out_file = out_dir / "JetBrainsMono-Italic.glyphs"
        assert out_file.is_file()

    def test_with_output_file(self):
        out_dir = get_tmp_path()
        out_dir.mkdir(exist_ok=True)
        out_file = out_dir / "custom.glyphs"
        package_to_single(get_mono_path(), out_file)
        assert out_file.is_file()

    @classmethod
    def tearDownClass(cls):
        rmdir(get_tmp_path())
